/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */

/* This device calculates activation time,              */
/* defined as time the value in the given layer         */
/* crossing the given threshold in the goven direction. */
/* The given field is in layer v0,			*/
/* the time is the value of main loop counter t,	*/
/* and the result is put in layer v1.			*/
/* An extra layer vd is used to remember the previous   */
/* value of the given field.				*/
/* The sign means: -1 down, +1 up, 0 either way		*/


#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#include "bikt.h"
#include "qpp.h"

typedef struct {
  real threshold;
  int sign;
  int vd;
} STR;

/****************/
RUN_HEAD(activation)
  DEVICE_CONST(real,threshold)
  DEVICE_CONST(int,sign)
  DEVICE_CONST(int,vd)
  int x, y, z;
  real y0, y1, p, q;
  real *u;
  int prev=DV*(vd-s.v0);
  int res=DV*(s.v1-s.v0);
  for(z=s.z0;z<=s.z1;z++) {
    for(y=s.y0;y<=s.y1;y++) {
      for(x=s.x0;x<=s.x1;x++) {
	u=New+ind(x,y,z,s.v0);
	y0=u[prev]-threshold;
	y1=u[0]-threshold;
	if( ((sign>=0)&&(y0<=0)&&(y1>0))	/* crossing in positive direction expected and happened */
	    || ((sign<=0)&&(y0>=0)&&(y1<0))) { 	/* crossing in negative direction expected and happened */
	  p=y1/(y1-y0); q=-y0/(y1-y0); 		/* linear interpolation coefficients */
	  assert(p>=0); assert(p<=1);
	  assert(q>=0); assert(q<=1);
	  u[res]=p*(t-1)+q*t;  			/* interpolate the t variable */
	}
	u[prev]=u[0];
      }
    }
  }
RUN_TAIL(activation)

DESTROY_HEAD(activation)
DESTROY_TAIL(activation)

CREATE_HEAD(activation)
  ACCEPTR(threshold,RNONE,RNONE,RNONE);
  ACCEPTI(sign,INONE,-1,1);
  ACCEPTI(vd,INONE,0,(int)vmax-1);
  ASSERT( dev->s.v1 != dev->s.v0 );
  ASSERT( S->vd != dev->s.v0 );
  ASSERT( S->vd != dev->s.v1 );
CREATE_TAIL(activation,1)

