/**
 * Experimental module, not for distribution. 
 * Interface with cardiac cell model description,
 * describing HH-type gates separately. 
 */

#ifndef _ionic
#define _ionic
typedef struct {		/* description of dependent parameters */
  int n;
  int *src;
  real **dst;
} Var;

#define IONICFTAB(name) int name(real V, real *values, int ntab)
typedef IONICFTAB(IonicFtab);
#define IONIC_FTAB_HEAD(name)	\
IONICFTAB(ftab_##name) {
#define IONIC_FTAB_TAIL(name)	\
  return 1;             	\
}

#define IONICFDDT(name) int name(real *u,int nv,real *values,int ntab,Par par,Var var,real *du,int no,real *nalp,real *nbet,int nn)
typedef IONICFDDT(IonicFddt);
/* Header of a standard ionic rhs calculator: */
/* - nostrify the parameters list, */
/* - check dimensions of subvectors, */
/* - implement parameter substitution if needed. */
#define IONIC_FDDT_HEAD(name,NV,NTAB,NO,NN)	\
IONICFDDT(fddt_##name) {	\
  STR *S = (STR *)par;	\
  int ivar;		\
  ASSERT(nv==NV);	\
  ASSERT(ntab==NTAB);	\
  ASSERT(no==NO);	\
  ASSERT(nn==NN);	\
  if (var.n) for(ivar=0;ivar<var.n;ivar++) *(var.dst[ivar])=u[var.src[ivar]];

#define IONIC_FDDT_TAIL(name)		\
  return 1;             \
}

/* Solver-independent entities exported by an ionic model description */
typedef struct {
  IonicFtab *ftab;		/* voltage dependent functions that can be tabulated */
  IonicFddt *fddt;		/* right-hand sides of non-gate equations */
  int no;			/* number of non-gate variables */
  int nn;			/* number of nontab gate variables */
  int nt;			/* number of tab gate variables */
  int ntab;			/* number of tabulated functions (number of columns in the table) */
  int V_index;			/* index of trmbr voltage in the state vector */
  Par p;			/* vector of model parameters */
  Var var;			/* description of dependent parameters */
} ionic_str;


#define IONICCREATE(name) int name(ionic_str *I,char *w,real **u,int v0)
typedef IONICCREATE(IonicCreate);

#define IONIC_CREATE_HEAD(name)				\
IONICCREATE(create_##name) {				\
  STR *S = (STR *)Calloc(1,sizeof(STR));        	\
  char *p=w;                                    	\
  Var *var=&(I->var);					\
  int ivar=0;						\
  if (!S) ABORT("cannot create %s",#name);		\
  for(var->n=0;*p;var->n+=(*(p++)==AT));		\
  if(var->n){CALLOC(var->dst,var->n,sizeof(real *));	\
  CALLOC(var->src,var->n,sizeof(int));}			\
  else {var->src=NULL;(var->dst)=NULL;}

#define IONIC_CREATE_TAIL(name,rc)    	\
  var->n=ivar;				\
  if(ivar){REALLOC(var->dst,1L*ivar*sizeof(real *));\
  REALLOC(var->src,1L*ivar*sizeof(int));}	\
  else{FREE(var->dst);FREE(var->src);}	\
  I->p = S;				\
  ASSERT(NV==NO+NN+NT);			\
  I->no = NO;				\
  I->nn = NN;				\
  I->nt = NT;				\
  I->ntab = NTAB;			\
  return rc;				\
}

#define IONIC_CONST(type,name) type name=S->I.name;


#endif
